from subprocess import run, PIPE
from typing import Dict, Optional


def build(substitutions: Optional[Dict[str, str]] = None) -> None:
    cmd = [
        "gcloud",
        "builds",
        "submit",
        ".",
        "--project=aaa-private-assets",
    ]
    if substitutions:
        substitutions_value = ",".join([f"{k}={v}" for k, v in substitutions.items()])
        cmd.append(f"--substitutions={substitutions_value}")
    print(
        "cloudbuild.yaml found, running build (https://console.cloud.google.com/cloud-build/builds?project=aaa-private-assets)..."
    )
    run(cmd, stdout=PIPE, stderr=PIPE)



from typing import Optional
from dataclasses import dataclass
from subprocess import run, PIPE
from .exceptions import LoggedError


@dataclass
class Commit:
    sha: str
    first_parent: Optional[str] = None
    second_parent: Optional[str] = None

    @property
    def is_empty_merge(self) -> bool:
        """Check if commit is a merge that could have been fast-forward."""
        if self.first_parent and self.second_parent:
            # If the first parent is an ancestor of the second, then the head branch
            # was not behind the base branch by any commits, meaning a fast-forward was possible
            return _is_ancestor(self.first_parent, self.second_parent)

        # Not a merge
        return False


class DirtyWorkingTreeError(LoggedError):
    def __init__(self) -> None:
        super().__init__("Working tree is dirty.")


class NoRepoError(LoggedError):
    def __init__(self) -> None:
        super().__init__("Not a git repository or no remote 'origin' set.")


def ensure_repo() -> None:
    cmd = ["git", "remote", "get-url", "origin"]
    result = run(cmd, stdout=PIPE, stderr=PIPE)
    if result.returncode != 0:
        raise NoRepoError()


def ensure_clean_working_tree() -> None:
    cmd = ["git", "status", "--short"]
    result = run(cmd, stdout=PIPE, stderr=PIPE)
    output = result.stdout.decode().strip()
    if len(output) > 0:
        raise DirtyWorkingTreeError()


def get_head() -> Commit:
    """Get the head commit with its parent(s)"""
    cmd = ["git", "log", "--pretty=%h %p", "-n", "1"]
    result = run(cmd, stdout=PIPE)
    shas = result.stdout.decode().strip().split(" ")
    return Commit(*shas)


def get_repo_name() -> str:
    cmd = ["git", "remote", "get-url", "origin"]
    result = run(cmd, stdout=PIPE)
    url = result.stdout.decode().strip()
    name = url.split("/")[-1].split(".")[0].replace("github_aaa-ncnu_", "")
    return name


def _is_ancestor(commit_a: str, commit_b: str) -> bool:
    """Checks if commit_a is an ancestor of commit_b"""
    cmd = ["git", "merge-head", "--is-ancestor", commit_a, commit_b]
    result = run(cmd, stdout=PIPE)
    return result.returncode == 0



import os
from argparse import ArgumentParser, Namespace

from . import git, plans
from .plans import PlanNotFound
from .build import build
from .terraform import Terraform
from .exceptions import LoggedError

TF_DIR = "tf"


def main() -> None:
    try:
        git.ensure_repo()

        parser = ArgumentParser()

        subparsers = parser.add_subparsers(title="actions")

        deployment_parser = ArgumentParser(add_help=False)
        deployment_parser.add_argument(
            "workspace",
            help="The Terraform workspace to select before planning or applying. Usually dev, qa, or prod.",
        )

        parser_plan = subparsers.add_parser(
            "plan",
            parents=[deployment_parser],
            help="Create and upload an execution plan based on the current head commit",
        )
        parser_plan.add_argument(
            "--refresh-only",
            action="store_true",
            help="Only refresh state",
        )
        parser_plan.set_defaults(func=plan)

        parser_apply = subparsers.add_parser(
            "apply",
            parents=[deployment_parser],
            help="Apply the execution plan for the current head commit",
        )
        parser_apply.set_defaults(func=apply)

        args = parser.parse_args()

        if not hasattr(args, "func"):
            parser.print_help()
            return

        args.func(args)

    except LoggedError as e:
        print(e)


def plan(args: Namespace):
    git.ensure_clean_working_tree()
    head = git.get_head()
    if os.path.exists("cloudbuild.yaml"):
        build(substitutions=dict(SHORT_SHA=head.sha))
    tf = Terraform(TF_DIR)
    tf.select_workspace(args.workspace)
    tf.plan(
        out=head.sha,
        vars=dict(commit_sha=head.sha),
        refresh_only=args.refresh_only,
    )
    planfile = _planfile(head.sha)
    plans.upload(args.workspace, planfile)
    os.remove(planfile)


def apply(args: Namespace) -> None:
    head = git.get_head()
    try:
        planfile = _planfile(head.sha)
        plans.download(args.workspace, planfile)
    except PlanNotFound:
        if head.second_parent and head.is_empty_merge:
            planfile = _planfile(head.second_parent)
            plans.download(args.workspace, planfile)
        else:
            raise
    tf = Terraform(TF_DIR)
    tf.select_workspace(args.workspace)
    tf.apply(os.path.basename(planfile))


def _planfile(sha: str) -> str:
    return os.path.join(TF_DIR, sha)
